#include <iostream>
#include <thread>
#include <queue>
#include <mutex>
#include <cstdint>
#include <condition_variable>

namespace
{

template<typename QueueItem>
class Queue
{
public:
    Queue() :
    	mQueue{},
		mMutex{},
		mCondVar{}
    {
    }

    void pushItem(const QueueItem& item)
    {
    	{
			std::lock_guard lk_guard{mMutex};
			mQueue.push(item);
    	}
        mCondVar.notify_one();
    }

    void popItem(std::stop_token stoken, QueueItem& item)
    {
        std::unique_lock lock{mMutex};

        const bool ret = mCondVar.wait(lock, stoken, [&]{return !mQueue.empty();});
        if(ret)
        {
            std::cout << "real notification\n";
        }
        else  /// wait returned due to 'stop'
        {
            std::cout << "stop request received\n";
            return;
        }

        item = mQueue.front();
        mQueue.pop();
    }

    bool popItem(std::stop_token stoken, QueueItem& item, uint32_t waitDurationMilliseconds)
    {
        std::unique_lock lock{mMutex};

        if (mCondVar.wait_for(lock, stoken, std::chrono::milliseconds(waitDurationMilliseconds),
				[&]{return !mQueue.empty();}) == true)
        {
            item = mQueue.front();
            mQueue.pop();

            return true;
        }
        /// due to time out or due to stop request
        if(stoken.stop_requested())
        {
            std::cout << "due to stop request\n";
        }
        else
        {
            std::cout << "due to timeout\n";
        }

        return false;
    }

private:
    std::queue<QueueItem> mQueue;
    mutable std::mutex mMutex;
    std::condition_variable_any mCondVar;
};

using Integers = Queue<int>;

/// we produce 3 times
void producer(Integers& qis)
{
	qis.pushItem(10);
	std::this_thread::sleep_for(std::chrono::seconds(3));
	qis.pushItem(20);
	std::this_thread::sleep_for(std::chrono::seconds(3));
	qis.pushItem(30);
	std::this_thread::sleep_for(std::chrono::seconds(6)); /// let's sleep some more so thread does not yet stop
}

/// we consume 3 times (exact match with the number of productions)
void consumer(std::stop_token stoken, Integers& qis)
{
	int value = 0;
	qis.popItem(stoken, value);
	std::cout << value << '\n';
	qis.popItem(stoken, value);
	std::cout << value << '\n';
	qis.popItem(stoken, value);
	std::cout << value << '\n';

	/// now we wait for a 4th consumption, which will never arrive; but main will stop us
	qis.popItem(stoken, value);
}

/// we consume 4 times, one time more than the number of productions ==> timeout for the last one
/// and we consume a 5th time --> we will see the 'stop requested'
void consumerTimeOut(std::stop_token stoken, Integers& qis)
{
	const uint32_t timeOutInMilliSeconds = 4000;
	int value = 0;
	if(qis.popItem(stoken, value, timeOutInMilliSeconds))
	{
		std::cout << value << '\n';
	}
	if(qis.popItem(stoken, value, timeOutInMilliSeconds))
	{
		std::cout << value << '\n';
	}
	if(qis.popItem(stoken, value, timeOutInMilliSeconds))
	{
		std::cout << value << '\n';
	}
	if(qis.popItem(stoken, value, timeOutInMilliSeconds))
	{
		std::cout << value << '\n';
	}
	if(qis.popItem(stoken, value, timeOutInMilliSeconds))
	{
		std::cout << value << '\n';
	}
}

} // namespace

int main()
{
	{
		Integers qis;
		std::jthread cons{consumer, std::ref(qis)};
		std::jthread prod{producer, std::ref(qis)}; /// this one dies first --> join on it
	}

	std::cout << "\n\n\n";

	{
		Integers qis;
		std::jthread cons{consumerTimeOut, std::ref(qis)};
		std::jthread prod{producer, std::ref(qis)};
	}

    return 0;
}
